/*******************************************************************
*
*  DESCRIPTION: Real Functions
*
*  AUTHOR: Daniel A. Rodriguez 
*
*  EMAIL: mailto://drodrigu@dc.uba.ar
*
*  DATE: 2/3/1999 (v2)
*
*******************************************************************/

/** include files **/
#include <stdlib.h>
#include "realfunc.h"
#include "mathincl.h"
extern "C"
{
	#include "randlib.h"
}


// Tolerance for the tangent
#define	TOL_TAN		7.55e-8

/* This function is only for Windows because it's not have the "rint" */
#ifdef __WINDOWS__
	double rint(double x)
	{
		double a1, fracc = modf( x, &a1 );
		return fracc < 0.5 ? floor(x) : ceil(x);
	}
#endif


Real tan (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	Real	w( remainder( r, Real(2 * M_PI) ) );	// Normalizo

	// Veo si se producira overflow
	if ( fabs(w.value() - M_PI / 2) <= TOL_TAN   ||   fabs(w.value() - M_PI * 3 / 2) <= TOL_TAN)
		return Real( M_INFINITO );

	return Real( tan( w.value() ) );
}

Real tanh (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	return Real( tanh( r.value() ) );
}

Real sec (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	Real w( cos( r ) );
	if (w.value() == 0)
		return Real( M_INFINITO );

	return Real( 1 / w.value() );
}

Real sech (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	return Real( 1 / cosh(r).value() );
}

Real sqrt (const Real &r)
{
	if ( r.IsUndefined() || r.value() < 0 )
		return Real();	// Devuelve un indefinido si es indefinido o negativo

	return Real( sqrt( r.value() ) );
}


Real sinh (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	return Real( sinh( r.value() ) );
}

Real sin (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	return Real( sin( r.value() ) );
}

Real cos (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	return Real( cos( r.value() ) );
}

Real cosh (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	return Real( cosh( r.value() ) );
}

Real trunc (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	return Real( floor( r.value() ) );
}

Real truncUpper (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	return Real( ceil( r.value() ) );
}

Real round (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	return Real( rint( r.value() ) );
}

Real fractional (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	double a1;
	return Real( modf( r.value(), &a1 ) );
}

Real remainder (const Real &r1, const Real &r2)
{
	if ( r1.IsUndefined() || r2.IsUndefined() || r2.value() == 0)
		return Real();	// Devuelve un indefinido

	return Real( fmod( r1.value(), r2.value() ) );
}

Real valueWithQuantum(Real &v, Real q)
{
	if (q.value() == 0)
		return v;

	double ss;

	modf(v.value()/q.value() + DBL_EPSILON, &ss);
					// Dejo en ss la part entera de v/q

	return Real( ss * q.value() );
}

Real power (const Real &r1, const Real &r2)
{
	if ( r1.IsUndefined() || r2.IsUndefined() )
		return Real();	// Devuelve un indefinido

	if ( r1.value() < 0 || fractional(r2).value() != 0)
		return Real();	// Si r1 es negativo o r2 no es entero

	return Real( pow( r1.value(), r2.value() ) );
}

Real max (const Real &r1, const Real &r2)
{
	if ( r1.IsUndefined() || r2.IsUndefined() )
		return Real::tundef;

	return Real( r1.value() > r2.value() ? r1.value() : r2.value() );
}

Real min (const Real &r1, const Real &r2)
{
	if ( r1.IsUndefined() || r2.IsUndefined() )
		return Real::tundef;

	return Real( r1.value() < r2.value() ? r1.value() : r2.value() );
}

Real abs (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	return Real( fabs( r.value() ) );
}

Real exp (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	return Real( exp( r.value() ) );
}

Real ln (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido
	if ( r.value() <= 0 )
		return Real();

	return Real( log( r.value() ) );
}

Real log (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido
	if ( r.value() <= 0 )
		return Real();

	return Real( log10( r.value() ) );
}

Real atanh (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido
	if ( fabs( r.value() ) > 1 )
		return Real();

	return Real( atanh( r.value() ) );
}

Real atan (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	return Real( atan( r.value() ) );
}

Real asinh (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	return Real( asinh( r.value() ) );
}

Real asin (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido
	if ( fabs (r.value()) > 1 )
		return Real();

	return Real( asin( r.value() ) );
}

Real acosh (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido
	if ( r.value() < 1 )
		return Real();

	return Real( acosh( r.value() ) );
}

Real acos (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido
	if ( fabs (r.value()) > 1 )
		return Real();

	return Real( acos( r.value() ) );
}

Real sign (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido
	if ( r.value() > 0 )
		return Real(1);
	if ( r.value() < 0 )
		return Real(-1);

	return Real( 0 );
}

TValBool isInt (const Real &r)
{
	if ( r.IsUndefined() )
		return TValBool::tfalse;

	Real	w( r - trunc(r) );

	if ( w.value() > 0)
		return TValBool::tfalse;

	return TValBool::ttrue;	
}

TValBool even (const Real &r)
{
	return isInt( r / Real(2) );
}

TValBool odd (const Real &r)
{
	if ( isInt(r) == TValBool::tfalse )
		return TValBool::tfalse;       // es indefinido o no es entero

	if ( even(r) == TValBool::ttrue )
		return TValBool::tfalse;

	return TValBool::ttrue;		// es entero y no es par
}

TValBool isPrime (const Real &r)
{
	if ( isInt(r) == TValBool::tfalse )
		return TValBool::tfalse;    // si es indefinido o no es entero

	Value	num = r.value(), aux = 3, aux2;

	if ( fmod( num, 2 ) == 0 )	// Si 2 divide a num
		return TValBool::tfalse;

	aux2 = sqrt(num);

	while ( aux <= aux2   &&   fmod( num, aux ) != 0 )
		aux += 2;

	return aux > aux2 ? TValBool::ttrue : TValBool::tfalse;
}

Real logn (const Real &v, const Real &n)
{
	if ( v.IsUndefined() || n.IsUndefined() )
		return Real();	// Devuelve un indefinido

	if (v.value() <= 0 || n.value() <= 0)
		return Real();

	return Real( log(v.value()) / log(n.value()) );
}

Real root (const Real &v, const Real &n)
{
	if ( v.IsUndefined() || n.IsUndefined() )
		return Real();	// Devuelve un indefinido
	if ( v.value() < 0 || n.value() == 0 )
		return Real();
	if ( v.value() == 0 )
		return Real(0);

	return Real( exp( (1 / n.value()) * log( v.value() ) ) );
}

Real fact (const Real &n)
{
	if ( isInt(n) == TValBool::tfalse )
		return Real();	// Devuelve un indefinido

	if ( n.value() < 0 )
		return Real();

	register long ff = 1;
	for (register int a=2; a <= n.value(); a++){
		ff *= a;
		
		if (ff < 0)	// Hubo Overflow
			return Real(M_INFINITO);
	}
	return Real(ff);
}

Real comb (const Real &m, const Real &k)
{
	// Aqui el combinatorio |m| = (m / k) * (m-1 / k-1) * ... (m-k+1 / 1)
	//                      |k|

	if ( isInt(m) == TValBool::tfalse || isInt(k) == TValBool::tfalse )
		return Real();	// Devuelve un indefinido

	if ( m.value() < k.value() || m.value() <= 0)
		return Real();

	register double aux = 1;

	for (register double w = 0; w <= k.value() - 1; w++){
		aux *= (m.value() - w) / (k.value() - w);

		if (aux < 0)		// Se produjo Overflow
			return Real(M_INFINITO);
	}
	return Real(aux);
}

Real random(const Real &r)
{
	if ( r.IsUndefined() )
		return Real::tundef;

	return uniform( Real::zero, Real::one );
}

Real chi(const Real &r)
{
	if ( r.IsUndefined() || r.value() <= 0 )
		return Real::tundef;

	return Real( genchi( r.value() ) );
}

Real beta(const Real &a, const Real &b)
{
	if ( a.IsUndefined() || b.IsUndefined() )
		return Real::tundef;
	if ( a.value() < 1e-37 || b.value() < 1e-37)
		return Real::tundef;

	return Real( genbet( a.value(), b.value() ) );
}

Real exponential(const Real &r)
{
	if ( r.IsUndefined() || r.value() < 0 )
		return Real::tundef;

	return Real( genexp( r.value() ) );
}

Real f(const Real &a, const Real &b)
{
	if ( a.IsUndefined() || b.IsUndefined() )
		return Real::tundef;
	if ( a.value() <= 0 || b.value() <= 0)
		return Real::tundef;

	return Real( genf( a.value(), b.value() ) );
}

Real gamma(const Real &a, const Real &b)
{
	if ( a.IsUndefined() || b.IsUndefined() )
		return Real::tundef;
	if ( a.value() <= 0 || b.value() <= 0)
		return Real::tundef;

	return Real( gengam( a.value(), b.value() ) );
}

Real normal(const Real &a, const Real &b)
{
	if ( a.IsUndefined() || b.IsUndefined() )
		return Real::tundef;
	if ( b.value() < 0)
		return Real::tundef;

	return Real( gennor( a.value(), b.value() ) );
}

Real uniform(const Real &a, const Real &b)
{
	if ( a.IsUndefined() || b.IsUndefined() )
		return Real::tundef;
	if ( a.value() > b.value() )
		return Real::tundef;

	return Real( genunf( a.value(), b.value() ) );
}

Real randInt(const Real &n)
{
	return round( uniform(0,n) );
}

Real binomial(const Real &n, const Real &p)
{
	if ( n.IsUndefined() || p.IsUndefined() )
		return Real::tundef;
	if ( n.value() < 0 || !isInt(n) )
		return Real::tundef;
	if ( p.value() > 1 || p.value() < 0)
		return Real::tundef;

	return Real( ignbin( (long) n.value(), p.value() ) );
}

Real poisson(const Real &m)
{
	if ( m.IsUndefined() || m.value() < 0 )
		return Real::tundef;

	return Real( ignpoi( m.value() ) );
}

Real mcm (const Real &r1, const Real &r2)
{
	if ( r1.IsUndefined() || r2.IsUndefined() )
		return Real::tundef;
	if ( !isInt(r1) || !isInt(r2) )
		return Real::tundef;

	register long	t, r, m, d, a = (long) r1.value(), b = (long) r2.value();

	if (b >= a)
		t=a, a=b, b=t;		// Swap(a,b)
	if ( (float) a / b == (long)(a/b))
		return Real(a);
	d = (long) a/b;
	do {
		d++;
		m = d * b;
		r = m - a * (long)(m/a);
	} while (r != 0);
	return Real(m);
}

Real gcd (const Real &r1, const Real &r2)
{
	if ( r1.IsUndefined() || r2.IsUndefined() )
		return Real::tundef;

	if ( !isInt(r1) || !isInt(r2) )
		return Real::tundef;

	register long	t, q, r, k = 0, a = (long) r1.value(), b = (long) r2.value();

	if (b >= a)
		t=a, a=b, b=t;		// Swap(a,b)

	while ( k == 0){
		q = (long) (a/b);
		r = (long) (a-b*q);
		if (r == 0)
			k = 1;
		a = b;
		b = r;
	}
	return Real(a);
}

Real hip (const Real &c1, const Real &c2)
{
	if ( c1.IsUndefined() || c2.IsUndefined() )
		return Real::tundef;

	if (c1.value() < 0 || c2.value() < 0)
		return Real::tundef;

	return Real( sqrt( c1.value() * c1.value() + c2.value() * c2.value() ) );
}

Real rectToPolar_r (const Real &x, const Real &y)
{
	Real c1( abs(x) ), c2( abs(y) );
	return hip(c1,c2);
}

Real rectToPolar_angle (const Real &x, const Real &y)
{
	return atan( y / x );
}

Real polarToRect_x (const Real &r, const Real &angle)
{
	if ( r.IsUndefined() || angle.IsUndefined() )
		return Real::tundef;

	if ( r.value() < 0 )
		return Real::tundef;

	return Real( r.value() * cos( angle.value() ) );
}

Real polarToRect_y (const Real &r, const Real &angle)
{
	if ( r.IsUndefined() || angle.IsUndefined() )
		return Real::tundef;

	if ( r.value() < 0 )
		return Real::tundef;

	return Real( r.value() * sin( angle.value() ) );
}

Real cotan (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	Real w( tan( r ) );
	if (w.value() == 0)
		return Real( M_INFINITO );

	return Real( 1 / w.value() );
}

Real cosec (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	Real w( sin( r ) );
	if (w.value() == 0)
		return Real( M_INFINITO );

	return Real( 1 / w.value() );
}

Real cosech (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	Real w( sinh( r ) );
	if (w.value() == 0)
		return Real( M_INFINITO );

	return Real( 1 / w.value() );
}

Real randomSign(const Real &r)
{
	if ( r.IsUndefined() )
		return Real::tundef;

	if ( uniform( Real(-1), Real(1) ).value() >= 0)
		return Real(1);

	return Real(-1);
}

Real degToRad(const Real &r)
{
	if ( r.IsUndefined() )
		return Real::tundef;

	return Real( r.value() * M_PI / 180 );
}

Real radToDeg(const Real &r)
{
	if ( r.IsUndefined() )
		return Real::tundef;

	return Real( r.value() * 180 / M_PI );
}

Real nextPrime (const Real &n)
{
	if ( n.IsUndefined() )
		return Real::tundef;

	Real	k( trunc( n ) );

	if (k == 2)
		return Real(3);
	else if (k < 2)
		return Real(2);

	// Sino k > 2
	if ( odd(k) == TValBool::ttrue )
		k = k + Real(2);
	else
		k = k + Real(1);
	
	while ( isPrime(k) != TValBool::ttrue   &&   k.value() < M_INFINITO )
		k = k + Real(2);
	
	return k;
}

Real nth_Prime (const Real &n)
{
	if ( n.IsUndefined() || !isInt(n) || n.value() < 1 )
		return Real::tundef;

	if (n.value() == 1)
		return Real(2);

	// Sino n >= 2
	Real	k = 3;
	register long	i = 2;

	while ( i < n.value() )
	{
		k = nextPrime( k );
		i++;

		if (k.value() == M_INFINITO)
			return Real(M_INFINITO);
	}
	return k;
}

Real realIf (const TValBool &c, const Real &t, const Real &f)
{
	if (c == TValBool::ttrue)
		return t;

	return f;
}

Real realIfu (const TValBool &c, const Real &t, const Real &f, const Real &u)
{
	if (c == TValBool::ttrue)
		return t;

	if (c == TValBool::tfalse)
		return f;

	return u;
}

Real asec (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido
	if ( fabs (r.value()) < 1 )
		return Real();

	return Real( atan ( r.value() / sqrt( r.value() * r.value() - 1) ) + sign( sign(r.value()) - 1) * M_PI/2 );
}

Real acotan (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	return Real( atan ( r.value() ) + M_PI / 2 );
}

Real asech (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	if ( r.value() == 0)
		return Real( M_INFINITO );

	return Real( log ( ( sqrt( (-r.value() * r.value() + 1) + 1 ) / r.value() ) ) );
}

Real acosech (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	if ( r.value() == 0)
		return Real( M_INFINITO );

	return Real( log ( ( sign(r.value()) * sqrt( (r.value() * r.value() + 1 ) + 1 ) / r.value() ) ) );
}

Real acotanh (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	if ( r.value() == 1)
		return Real( M_INFINITO );

	return Real( log ( (r.value() + 1) / (r.value() - 1) ) / 2 );
}

Real CtoF (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	return Real( (9 * r.value() / 5) + 32);
}

Real CtoK (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	return Real( r.value() + 273 );
}

Real KtoC (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	return Real( r.value() - 273 );
}

Real KtoF (const Real &r)
{
	return CtoF( KtoC(r) );
}

Real FtoC (const Real &r)
{
	if ( r.IsUndefined() )
		return Real();	// Devuelve un indefinido

	return Real( (r.value() - 32) * 5 / 9 );
}

Real FtoK (const Real &r)
{
	return CtoK( FtoC(r.value()) );
}

TValBool isUndefinedReal(const Real &r)
{
	if ( r.IsUndefined() )
		return TValBool::ttrue;

	return TValBool::tfalse;
}


/********************************
*LSM-Prueba funcion - Hodgkin-Huxley Model of an Active Excitable Membrane
********************************/
Real AP (const Real &r1, const Real &r2)
{
#include <math.h>
#include <stdio.h>

/* parameters of the membrane */
double  Cm	=1.0; //1.0	/*uF/cm2, membrane capacitance */
double  Rm	=1.0;		/*kOhm/cm2, membrane resistance */
double 	gna	=16.0;   // 120
double 	gk	=36.0;	//36.0
double 	gl	=0.3;
double 	Ena	=50.00;	//57.56
double 	Ek	=-96.00;	//-72.26
double 	El	=262.0;	//-49.387
double 	Vrest=	-60.094;   //-60.094
double  tend=	9.0;
double 	dt	=0.01;

double Vm,Vmnew,t,stim,Ina,Ik,Il,mnew,nnew,hnew;
double an,am,ah,bn,bm,bh,vm,Is,n,m,h,dur,Vresti,tsim;
static double ns[10][10],ms[10][10],hs[10][10],durs[10][10],tsims[10][10],Vrestis[10][10];
int i,j;
double e;
i=r1.value()/10000;
j=(r1.value()-i*10000)/100;
e=r1.value()-(i*10000+j*100);
Vresti=r2.value();

dur = 0.1 ; //dur = 0.1;
n=ns[i][j];
m=ms[i][j];
h=hs[i][j];
tsim=tsims[i][j];

if(e<0.5 && e>0)
{
	tsim=10;
	Vresti=Vrest;
}
else
{
	tsim=tsims[i][j]+10;
}


Is = -80.0 ;  //Is = -70.0;
Vm = Vresti;
vm = Vm- Vrest;
am = (0.1*(25-vm))/(exp((25-vm)/10) -1);
an = (.01*(10-vm))/(exp((10-vm)/10) -1);
ah= 0.07*(exp(-vm/20));
bm = 4*(exp(-vm/18));
bn = .125*(exp(-vm/80));
bh =1/(exp((30-vm)/10) +1) ;

if (Vresti==Vrest){
n = an/(an+bn);
m = am/(am+bm);
h = ah/(ah+bh);
n = n + dt*((an*(1-n)) - (bn*n));
m = m + dt*((am*(1-m)) - (bm*m));
h = h + dt*((ah*(1-h)) - (bh*h));
}

/* apply stimulating current */
t=-(11-((tsim)/10))/100;

if ((t>=0.0) && (t<=dur)) stim = Is; else stim = 0.0;
/* new current stuff */
Ina = gna*h*m*m*m*(Vm - Ena);
Ik = gk*n*n*n*n*(Vm- Ek);
Il = gl*(Vm- El);
/* updating dependent variable: Vm  and vm*/
Vm = Vm + (dt/Cm)*(stim - Ina -Ik - Il);
vm = Vm- Vrest;

/* calculate alphas and betas  */
am = (0.1*(25-vm))/(exp((25-vm)/10) -1);
an = (.01*(10-vm))/(exp((10-vm)/10) -1);
ah= 0.07*(exp(-vm/20));
bm = 4*(exp(-vm/18));
bn = .125*(exp(-vm/80));
bh =1/(exp((30-vm)/10) +1) ;

/* compute n , m, and h  for next iteration*/
n = n + dt*((an*(1-n)) - (bn*n));
m = m + dt*((am*(1-m)) - (bm*m));
h = h + dt*((ah*(1-h)) - (bh*h));

ns[i][j]=n;
ms[i][j]=m;
hs[i][j]=h;
tsims[i][j]=tsim;

return Real(Vm);

}

/********************************
*LSM-Prueba funcion - Action Potential Model of a Human Atrial Cell
********************************/
Real APA (const Real &r1, const Real &r2)
{
#include <math.h>
#include <stdio.h>
/*Global Variables */
int i,ci,cj;
double VVecinos, Vant, Vrest;
double R,T,F,Cm,xVcell,Vi,Vup,Ko,Nao,Cao;
double gna,gk1,gto,gkr,gks,gcal,gbca,gbna,gkur;
double Inakmax,Inacamax,Ipcamax,Iupmax,Kq10,lambda;
double Kmnai,Kmko,Kmna,Kmca,ksat,krel,kup;
double Caupmax,Cmdnmax,Trpnmax,Csqnmax,KmCmdn,KmTrpn,KmCsqn;
double Iion,Ist,Ina,Ik1,Ito,Ikur,Ikr,Iks,Ical,Iup;
double Ipca,Inak,Inaca,Ibna,Ibca,Iupleak,Irel,Itr;
double dt,B1,B2;
double time,Vrel;
double am,bm,aj,bj,ah,bh,taum,tauh,tauj,infm,infh,infj;
double asa,bsa,asi,bsi,tausa,infsa,tausi,infsi;
double aui,bui,aua,bua,tauui,infui,tauua,infua;
double axr,bxr,infxr,tauxr,axs,bxs,tauxs,infxs;
double taud,infd,tauf,inff,taufca,inffca,fnak,sigma;
double tauu,infu,tauv,infv,tauw,infw,Fn,tautr;
double Ena,Ek,Eca;//Fn;

double V,Nai,Ki,Cai,Caup,Carel,m,h,j,sa,si;
double ua,ui,xr,xs,d,f,fca,u,v,w;

static double Nais[10][10],Kis[10][10],Cais[10][10],Caups[10][10],Carels[10][10];
static double ms[10][10],hs[10][10],js[10][10],sas[10][10],sis[10][10];
static double uas[10][10],uis[10][10],xrs[10][10],xss[10][10],ds[10][10];
static double fs[10][10],fcas[10][10],us[10][10],vs[10][10],ws[10][10],times[10][10];

V=r2.value();
ci=r1.value()/10000;
cj=(r1.value()-ci*10000)/100;
VVecinos=r1.value()-(ci*10000+cj*100);

Nai=Nais[ci][cj];
Ki=Kis[ci][cj];
Cai=Cais[ci][cj];
Caup=Caups[ci][cj];
Carel=Carels[ci][cj];
m=ms[ci][cj];
h=hs[ci][cj];
j=js[ci][cj];
sa=sas[ci][cj];
si=sis[ci][cj];
ua=uas[ci][cj];
ui=uis[ci][cj];
xr=xrs[ci][cj];
xs=xss[ci][cj];
d=ds[ci][cj];
f=fs[ci][cj];
fca=fcas[ci][cj];
u=us[ci][cj];
v=vs[ci][cj];
w=ws[ci][cj];
time=times[ci][cj];


/* constants */
dt = 0.005;		/*ms*/

R = 8.3143;  		/*J/mol*K*/
T = 310.0;		/*K*/
F = 96.4867;		/*C/mmol*/
Cm = 100.0;		/*pF*/
xVcell = 20100.0;	/*um^3*/
Vi = 13668.0;		/*um^3*/
Vup = 1109.52;		/*um^3*/
Vrel = 96.48;		/*um^3*/
Ko = 5.4;		/*mM*/
Nao = 140;		/*mM*/
Cao = 1.8;		/*mM*/
gna = 7.8*Cm;		/*nS/pF*/
gk1 = 0.09*Cm;		/*nS/pF*/
gto = 0.1652*Cm;		/*nS/pF*/
gkr = 0.0294*Cm;		/*nS/pF*/
gks = 0.129*Cm;		/*nS/pF*/
gcal = 0.1238*Cm;		/*nS/pF*/
gbca = 0.00113*Cm; 	/*nS/pF*/
gbna = 0.000674*Cm;	/*nS/pF*/
Inakmax = 0.6*Cm;		/*pA/pF*/
Inacamax = 1600.0*Cm;		/*pA/pF*/
Ipcamax = 0.275*Cm;	/*pA/pF*/
Iupmax = 0.005;		/*mM/ms*/
Kq10 = 3.0;
lambda = 0.35;
Kmnai = 10.0;		/*mM*/
Kmko = 1.5;		/*mM*/
Kmna = 87.5;		/*mM*/
Kmca = 1.38;		/*mM*/
ksat = 0.1;
krel = 30.0;		/*ms^-1*/
kup = 0.00092;		/*mM*/
Caupmax = 15.0;		/*mM*/
Cmdnmax = 0.05;		/*mM*/
Trpnmax = 0.07;		/*mM*/
Csqnmax = 10.0;		/*mM*/
KmCmdn = 0.00238;	/*mM*/
KmTrpn = 0.0005;	/*mM*/
KmCsqn = 0.8;		/*mM*/
taufca = 2.0;
tautr = 180.0;
tauu = 8.0;
Vrest=-82.2;
//Guardo el voltrage con el que entr...
Vant=V;
/* Initial Conditions */
//Si esta en el voltage de inicio inicializa las variables.
if (V > Vrest-1.0 && V < Vrest+1.0 )
{
	V = Vrest+1.0;
	h = 0.965;
	d = 1.37e-4;
	xr = 3.29e-5;
	Nai = 11.2;
	Ki = 139.0;
	Carel = 1.49;
	si = 0.999;
	ui = 0.999;
	v = 1.0;
	m = 2.91e-3;
	j = 0.978;
	f = 0.999;
	xs = 1.87e-2;
	Cai = 1.02e-4;
	Caup = 1.49;
	sa = 3.04e-2;
	ua = 4.96e-3;
	fca = 0.775;
	u = 0.0;
	w = 0.999;
	time=0.0;
}

/* initial calculations */
Ena = (R*T/F)*log(Nao/Nai);
Ek = (R*T/F)*log(Ko/Ki);
Eca = (R*T/(2.0*F))*log(Cao/Cai);
gkur = Cm*(0.005 + 0.05/(1.0+exp(-1.0*(V-15.0)/13.0)));
sigma = (1.0/7.0)*(exp(Nao/67.3) - 1.0);
fnak = 1.0/(1.0 + 0.1245*exp(-0.1*F*V/(R*T)) + 0.0365*sigma*exp(-1.0*F*V/(R*T)));

/* update currents */
Ina = gna*m*m*m*h*j*(V-Ena);
Ik1 = gk1*(V-Ek)/(1+exp(0.07*(V+80.0)));
Ito = gto*sa*sa*sa*si*(V-Ek);
Ikur = gkur*ua*ua*ua*ui*(V-Ek);
Ikr = gkr*xr*(V-Ek)/(1.0 + exp((V+15.0)/22.4));
Iks = gks*xs*xs*(V-Ek);
Ical = gcal*d*f*fca*(V-65.0);
Inak = Inakmax*fnak*(1.0/(1.0+ pow(Kmnai/Nai,1.5))) * Ko/(Ko+Kmko);
Inaca = Inacamax*(exp(lambda*F*V/(R*T))*Nai*Nai*Nai*Cao - exp((lambda-1)*F*V/(R*T))*Nao*Nao*Nao*Cai)/
((Kmna*Kmna*Kmna + Nao*Nao*Nao)*(Kmca + Cao)*(1 + ksat*exp((lambda-1)*F*V/(R*T))));
Ibca = gbca*(V-Eca);
Ibna = gbna*(V-Ena);
Ipca = Ipcamax*Cai/(0.0005 + Cai);

Irel = krel*u*u*v*w*(Carel - Cai);
Itr = (Caup-Carel)/tautr;
Iup = Iupmax/(1.0+(kup/Cai));
Iupleak = Caup*Iupmax/Caupmax;

if ((time>=0.0) && (time<=2.0)) Ist = -2000.0; else Ist = 0.0;

Iion=Ina+Ik1+Ito+Ikur+Ikr+Iks+Ical+Ipca+Inak+Inaca+Ibna+Ibca+Ist;

/* update variables */
Ena = (R*T/F)*log(Nao/Nai);
Ek = (R*T/F)*log(Ko/Ki);
Eca = (R*T/(2.0*F))*log(Cao/Cai);
Fn = 1e-12*Vrel*Irel - (5e-13/F)*(0.5*Ical - 0.2*Inaca);

if(V==-47.13){
am = 3.2;
}else{
am = 0.32*(V+47.13)/(1-exp(-0.1*(V+47.13)));
}
bm = 0.08*exp(-V/11.0);
taum = 1.0/(am+bm);
infm = am*taum;


if(V>=-40.0){
ah = 0.0;
}else{
ah = 0.135*exp( -1.0*(V+80.0)/6.8);
}
if(V>=-40.0){
bh = 1.0/(0.13*(1+exp(-1.0*(V+10.66)/11.1)));
}else{
bh = 3.56*exp(0.079*V) + (3.1e5)*exp(0.35*V);
}
tauh = 1.0/(ah+bh);
infh = ah*tauh;

if(V>=-40.0){
aj = 0.0;
}else{
aj = ( -127140.0*exp(0.2444*V) - (3.474e-5)*exp(-0.04391*V) )*(V+37.78)/
	(1.0+exp(0.311*(V+79.23)));
}
if(V>=-40.0){
bj = 0.3*exp((-2.535e-7 )*V)/(1.0+exp(-0.1*(V+32.0)));
}else{
bj = 0.1212*exp(-0.01052*V)/(1.0+exp(-0.1378*(V+40.14)));
}
tauj = 1.0/(aj+bj);
infj = aj*tauj;

asa = 0.65/( (exp(-1.0*(V+10.0)/8.5)) + (exp(-1.0*(V-30.0)/59.0)) );
bsa = 0.65/(2.5 + exp((V+82.0)/17.0) );
infsa = 1.0/(1.0 + exp(-1.0*(V+20.47)/17.54));
tausa = (1.0/(asa+bsa))/Kq10;

asi = 1.0/(18.53 + exp((V+113.7)/10.95));
bsi = 1.0/(35.56 + exp(-1.0*(V+1.26)/7.44));
infsi = 1.0/(1.0 + exp((V+43.1)/5.3));
tausi = (1.0/(asi+bsi))/Kq10;


gkur = Cm*(0.005 + 0.05/(1.0+exp(-1.0*(V-15.0)/13.0)));
aua = 0.65*1.0/(exp(-1.0*(V+10.0)/8.5) + exp(-1.0*(V-30.0)/59.0));
bua = 0.65*1.0/(2.5+exp((V+82.0)/17.0));
infua = 1.0/(1.0 + exp(-1.0*(V+30.3)/9.6)); 
tauua = (1.0/(aua+bua))/Kq10;


aui = 1.0/(21.0 + exp(-1.0*(V-185.0)/28.0));
bui = exp((V-158.0)/16.0);
infui = 1.0/(1.0 + exp((V-99.45)/27.48)); 
tauui = (1.0/(aui+bui))/Kq10;

axr = 0.0003*(V+14.1)/(1.0 - exp(-1.0*(V+14.1)/5.0));
bxr = (7.3898e-5)*(V-3.3328)/(exp((V-3.3328)/5.1237)-1.0);
infxr = 1.0/(1.0 + exp(-1.0*(V+14.1)/6.5)); 
tauxr = (1.0/(axr+bxr))/Kq10;

axs = (4e-5)*(V-19.9)/(1.0 - exp(-1.0*(V-19.9)/17.0));
bxs = (3.5e-5)*(V-19.9)/(exp((V-19.9)/9.0) - 1.0);
tauxs = 0.5/(axs+bxs);
infxs = 1.0/sqrt(1.0 + exp(-1.0*(V-19.9)/12.7));

taud = (1.0 - exp(-1.0*(V+10.0)/6.24))/(0.035*(V+10.0)*(1.0+exp(-1.0*(V+10.0)/6.24)));
infd = 1.0/(1.0 + exp(-1.0*(V+10.0)/8.0));

tauf = 9.0/(0.0197*exp(-1.0*0.0337*0.0337*(V+10.0)*(V+10.0)) + 0.02);
inff = 1.0/(1.0 + exp((V+28.0)/6.9));

inffca = 1.0/(1.0+Cai/0.00035);

sigma = (1.0/7.0)*(exp(Nao/67.3) - 1.0);
fnak = 1.0/(1.0 + 0.1245*exp(-0.1*F*V/(R*T)) + 0.0365*sigma*exp(-1.0*F*V/(R*T)));

infu = 1.0/(1.0+ exp(-1.0*(Fn - 3.4175e-13)/13.67e-16));

tauv = 1.91 + 2.09/(1.0 + exp(-1.0*(Fn - 3.4175e-13)/13.67e-16));
infv = 1.0 - 1.0/(1.0 + exp(-1.0*(Fn - 6.835e-14)/13.67e-16));

tauw = 6.0*(1.0 - exp(-1.0*(V-7.9)/5.0))/( (1.0 + 0.3*exp(-1.0*(V-7.9)/5.0))*(V-7.9) );
infw = 1.0 - 1.0/(1.0 + exp(-1.0*(V-40.0)/17.0));


/* update differential equations */
V += dt*-1.0*Iion/Cm;
Nai += dt*( (-3.0*Inak - 3.0*Inaca - Ibna - Ina)/(F*Vi) );
Ki += dt*( (2.0*Inak-Ik1-Ito-Ikur-Ikr-Iks)/(F*Vi) );
B1 = (2.0*Inaca - Ipca - Ical - Ibca)/(2.0*F*Vi) + (Vup*(Iupleak-Iup) + Irel*Vrel)/Vi;
B2 = 1.0 + Trpnmax*KmTrpn/((Cai +KmTrpn)*(Cai +KmTrpn)) + Cmdnmax*KmCmdn/((Cai +KmCmdn)*(Cai +KmCmdn));
Cai += dt*(B1/B2);
Caup += dt*(Iup-Iupleak - Itr*Vrel/Vup);
Carel += dt*( (Itr-Irel)*(1.0/( 1.0 + (Csqnmax*KmCsqn)/( (Carel+KmCsqn)*(Carel+KmCsqn)))));

m += dt*(infm - m)/taum;
h += dt*(infh - h)/tauh;
j += dt*(infj - j)/tauj;
sa += dt*(infsa - sa)/tausa;
si += dt*(infsi - si)/tausi;
ua += dt*(infua - ua)/tauua;
ui += dt*(infui - ui)/tauui;
xr += dt*(infxr - xr)/tauxr;
xs += dt*(infxs - xs)/tauxs;
d += dt*(infd - d)/taud;
f += dt*(inff - f)/tauf;
fca += dt*(inffca - fca)/taufca;
u += dt*(infu - u)/tauu;
v += dt*(infv - v)/tauv;
w += dt*(infw - w)/tauw;

i = time/dt;

Nais[ci][cj]= Nai;
Kis[ci][cj]= Ki;
Cais[ci][cj]= Cai;
Caups[ci][cj]= Caup;
Carels[ci][cj]= Carel;
ms[ci][cj]= m;
hs[ci][cj]= h;
js[ci][cj]= j;
sas[ci][cj]= sa;
sis[ci][cj]= si;
uas[ci][cj]= ua;
uis[ci][cj]= ui;
xrs[ci][cj]= xr;
xss[ci][cj]= xs;
ds[ci][cj]= d;
fs[ci][cj]= f;
fcas[ci][cj]= fca;
us[ci][cj]= u;
vs[ci][cj]= v;
ws[ci][cj]= w;
times[ci][cj]= time+dt;

//Si esta llegando al voltaje de reposo estando en la repolarizacion 
//y tiene vecinos con corriente positiva=>comienza el nuevo AP inicializando el Voltaje
if(VVecinos >= 0.5 && VVecinos <= 1.5 && ( (V < Vrest+3.5 && V < Vant) || (V < Vrest-3.5) || (r2==Real())) ) V=Vrest;

return Real(V);
}
